from uuid import uuid4
import csv
from dataclasses import dataclass
from io import StringIO
from typing import Dict, List, Literal, Optional, Union, cast

import orjson
from sand.models.project import Project
from sand.models.table import Link, Table, TableRow
from werkzeug.datastructures import FileStorage
from sand.models import db


@dataclass
class CSVParserOpts:
    format: Literal["csv"] = "csv"
    delimiter: str = ","
    # whether the first row in header
    first_row_is_header: bool = True


@dataclass
class JSONParserOpts:
    format: Literal["json"] = "json"


ParserOpts = Union[CSVParserOpts, JSONParserOpts]


@dataclass
class RawTable:
    name: str
    header: List[str]
    # each item in the outer list of `rows` and `links` is for a row
    rows: List[List[Union[str, int, float]]]
    links: List[Dict[str, List[Link]]]


@dataclass
class UploadingTable:
    parser_opts: ParserOpts
    tables: List[RawTable]


ALLOWED_EXTENSIONS = {"json", "csv", "tsv"}


def get_extension(filename: str) -> Optional[str]:
    lst = filename.rsplit(".", 1)
    if len(lst) == 1:
        return None
    return lst[1].lower()


def save_upload(project: Project, raw_tables: List[RawTable]) -> List[Table]:
    """Save the upload results to the database"""
    with db.transaction():
        cursor = Table.select(Table.name).where(
            (Table.project == project) & (Table.name.in_([x.name for x in raw_tables]))
        )
        existing_names = {x.name for x in cursor}

        tables = []
        for raw_table in raw_tables:
            if raw_table.name not in existing_names:
                name = raw_table.name
            else:
                name = f"{raw_table.name}-{str(uuid4()).replace('-', '')}"

            table = Table(
                name=name,
                description="",
                columns=raw_table.header,
                project=project,
                size=len(raw_table.rows),
                context_values=[],
                context_tree=[],
            )
            table.save()
            tables.append(table)

        table_rows = []
        for raw_table, table in zip(raw_tables, tables):
            for i, row in enumerate(raw_table.rows):
                table_rows.append(
                    TableRow(table=table, index=i, row=row, links=raw_table.links[i])
                )

        TableRow.bulk_create(table_rows, batch_size=200)
        return tables


def parse_upload(
    user_preferred_parser_opts: Optional[ParserOpts], file: FileStorage
) -> List[UploadingTable]:
    assert file.filename is not None

    name = file.filename.rsplit(".", 1)[0]
    if user_preferred_parser_opts is not None:
        parser_opts = user_preferred_parser_opts
    else:
        ext = get_extension(file.filename)
        if ext in {"csv", "tsv"}:
            delimiter = "," if ext == "csv" else "\t"
            parser_opts = CSVParserOpts(format="csv", delimiter=delimiter)
        elif ext == "json":
            parser_opts = JSONParserOpts(format="json")
        else:
            raise ValueError(f"Invalid format: {ext}")

    if isinstance(parser_opts, CSVParserOpts):
        return parse_csv_file(name, file, parser_opts)

    if isinstance(parser_opts, JSONParserOpts):
        return parse_json_file(name, file, parser_opts)

    raise NotImplementedError()


def parse_csv_file(name: str, file: FileStorage, parser_opts: CSVParserOpts):
    reader = csv.reader(StringIO(file.read().decode()), delimiter=parser_opts.delimiter)
    rows = list(reader)

    n_columns = max(len(row) for row in rows)

    if parser_opts.first_row_is_header:
        header, rows = rows[0], rows[1:]
        if len(header) < n_columns:
            for i in range(len(header), n_columns):
                header.append(f"")
    else:
        header = [""] * n_columns

    for row in rows:
        if len(row) < n_columns:
            row.extend([""] * (n_columns - len(row)))

    return [
        UploadingTable(
            parser_opts=parser_opts,
            tables=[
                RawTable(
                    name=name,
                    header=header,
                    rows=cast(List[List[Union[str, int, float]]], rows),
                    links=[{} for _ in rows],
                )
            ],
        )
    ]


def parse_json_file(name: str, file: FileStorage, parser_opts: JSONParserOpts):
    records = orjson.loads(file.read())
    if not isinstance(records, list):
        raise Exception("JSON file must contain a list of rows")

    # now normalize and check the content & construct schema
    attrs = {}
    for record in records:
        for key, value in record:
            if key not in attrs:
                attrs[key] = len(attrs)

            if value is not None and not isinstance(value, (str, int, float)):
                raise Exception(
                    f"Invalid value type for attribute {key}. Expect string, number, or null"
                )

    rows = []
    header = list(attrs.keys())
    for record in records:
        rows.append([record.get(key, None) for key in attrs])

    return [
        UploadingTable(
            parser_opts=parser_opts,
            tables=[
                RawTable(name=name, header=header, rows=rows, links=[{} for _ in rows])
            ],
        )
    ]



import sys
import traceback
from typing import Dict, List, Callable, Any, Union, Iterable, Tuple, Optional, Literal
from typing_extensions import NotRequired, TypedDict
from dataclasses import dataclass

from flask import jsonify, request
from flask.blueprints import Blueprint
from RestrictedPython import compile_restricted_function, safe_globals
from werkzeug.exceptions import BadRequest

from sand.models.table import Link, Table, TableRow
from gena.deserializer import get_dataclass_deserializer
from gena import generate_api
from sand.models import Transformation

transformation_bp = generate_api(Transformation)


@dataclass
class Context:
    """Context dataclass to access the row of the cell that is being transformed."""

    index: int
    row: List[Union[str, float]]


@dataclass
class TransformRequestPayload:
    """Request Payload dataclass to validate the request obtained from the API call"""

    type: Literal["map", "filter", "split", "concatenate"]
    table_id: int
    mode: str
    datapath: Union[str, List[str]]
    code: str
    tolerance: int
    rows: Optional[int] = None
    outputpath: Optional[Union[str, List[str]]] = None


transform_request_deserializer = get_dataclass_deserializer(TransformRequestPayload, {})


class Tdata(TypedDict):
    path: int
    value: Union[str, List[str]]
    ok: NotRequired[Union[List, int, str, Iterable]]
    error: NotRequired[str]


def filter_traceback_errors() -> str:
    """Filters traceback errors, removes sensitive information

    Args:

    Returns:
        Error String without the sensitive information.
    """
    (exc, value, tb) = sys.exc_info()
    tb = tb.tb_next
    return "".join(traceback.format_exception(exc, value, tb))


Item = Any
ItemIndex = int


def transform_map(
    transform_func: Callable[[Any, Context], Any],
    data: Iterable[Tuple[ItemIndex, Item, Context]],
    tolerance: int,
) -> List[Tdata]:
    """Implements map transform, performs map operation over each cell, for a given column

    Args:
        transform_func: User defined python function defined by the user
        data: iterable with Column data and context object
        tolerance: contains the API request data

    Returns:
        list of Tdata objects, data transformed after applying map transform
    """
    transformed_data = []
    for path, value, context in data:
        value = value[0]
        tdata = Tdata(path=path, value=value)
        try:
            result = transform_func(value, context)
            tdata["ok"] = result
        except Exception as err:
            tdata["error"] = filter_traceback_errors()
            tolerance -= 1
            if tolerance == 0:
                transformed_data.append(tdata)
                return transformed_data
        transformed_data.append(tdata)
    return transformed_data


def transform_filter(
    transform_func: Callable[[Any, Context], Any],
    data: Iterable[Tuple[ItemIndex, Item, Context]],
    tolerance: int,
) -> List[Tdata]:
    """Implements filter transform, performs filter operation over each cell, for a given column

    Args:
        transform_func: User defined python function defined by the user
        data: iterable with Column data and context object
        tolerance: contains the API request data

    Returns:
        list of Tdata objects, data transformed after applying filter transform

    Raises:
        BadRequest: An error occurred when the transform_func on execution, does not return a boolean
    """
    transformed_data = []
    for path, value, context in data:
        value = value[0]
        tdata = Tdata(path=path, value=value)
        try:
            result = transform_func(value, context)
            if not isinstance(result, bool):
                raise BadRequest("filter transform function must return boolean value")
            tdata["ok"] = result
        except Exception as err:
            tdata["error"] = filter_traceback_errors()
            tolerance -= 1
            if tolerance == 0:
                transformed_data.append(tdata)
                return transformed_data
        transformed_data.append(tdata)
    return transformed_data


def transform_split(
    transform_func: Callable[[Any, Context], Any],
    data: Iterable[Tuple[ItemIndex, Item, Context]],
    tolerance: int,
) -> List[Tdata]:
    """Implements split transform, performs split operation over each cell, for a given column

    Args:
        transform_func: User defined python function defined by the user
        data: iterable with Column data and context object
        tolerance: contains the API request data

    Returns:
        list of Tdata objects, data transformed after applying split transform

    Raises:
        BadRequest: An error occurred when transform_func on execution, does not return a list
    """
    transformed_data = []
    for path, value, context in data:
        value = value[0]
        tdata = Tdata(path=path, value=value)
        try:
            result = transform_func(value, context)
            if not isinstance(result, Iterable):
                raise BadRequest("split transform function must return list")
            tdata["ok"] = result
        except Exception as err:
            tdata["error"] = filter_traceback_errors()
            tolerance -= 1
            if tolerance == 0:
                transformed_data.append(tdata)
                return transformed_data
        transformed_data.append(tdata)
    return transformed_data


def transform_concatenate(
    transform_func: Callable[[Any, Context], Any],
    data: Iterable[Tuple[ItemIndex, Item, Context]],
    tolerance: int,
) -> List:
    """Implements concatenate transform, performs concatenate operation over each cell, for a given column

    Args:
        transform_func: User defined python function defined by the user
        data: iterable with Column data and context object
        tolerance: contains the API request data

    Returns:
        list of Tdata objects, data transformed after applying concatenate transform
    """
    transformed_data = []
    for path, value, context in data:
        tdata = Tdata(path=path, value=value)
        try:
            result = transform_func(value, context)
            tdata["ok"] = result
        except Exception as err:
            tdata["error"] = filter_traceback_errors()
            tolerance -= 1
            if tolerance == 0:
                transformed_data.append(tdata)
                return transformed_data
        transformed_data.append(tdata)
    return transformed_data


# overriding inbuilt _getitem_ guard function
def custom_getitem_guard(obj: Any, index: int) -> Any:
    """Implements __getitem__ restrictedpython policy and wraps _getitem_ function

    Args:
        obj: object that has __getitem__ implementation in python
        index: index of the element that can be accessed from obj

    Returns:
        object element at position index
    """
    return obj[index]


def compile_function(code: str) -> Callable:
    """Executes code in string in a restricted mode using restrictedpython

    Args:
        code: object that has __getitem__ implementation in python

    Returns:
        Callable function that wraps the code as a function body

    Raises:
        BadRequest: An error occurred when the code has compilation error
    """
    loc = {}
    safe_globals.update({"_getitem_": custom_getitem_guard})
    compiled_result = compile_restricted_function("value,context", code, "<function>")

    if compiled_result.errors:
        raise BadRequest("\n".join(compiled_result.errors))

    exec(compiled_result.code, safe_globals, loc)

    return loc["<function>"]


@transformation_bp.route(f"/{transformation_bp.name}/test", methods=["POST"])
def transform():
    if isinstance(request.json["datapath"], str):
        request.json["datapath"] = [request.json["datapath"]]

    request_data = transform_request_deserializer(request.json)
    table = Table.get_by_id(request_data.table_id)
    table_rows: List[TableRow] = list(
        TableRow.select().where(TableRow.table == table).order_by(TableRow.index)
    )
    transform_func = compile_function(request_data.code)
    col_index_list = [table.columns.index(column) for column in request_data.datapath]
    data = (
        (
            table_row.index,
            [table_row.row[col_index] for col_index in col_index_list],
            Context(index=table_row.index, row=table_row.row),
        )
        for table_row in table_rows[: request_data.rows]
    )

    transformed_data = None

    if request_data.type == "map":
        if request_data.outputpath and len(request_data.outputpath) != 1:
            raise BadRequest(
                "For transform type map the outputpath should be a single column"
            )
        transformed_data = transform_map(transform_func, data, request_data.tolerance)

    elif request_data.type == "filter":
        if request_data.outputpath and len(request_data.outputpath) != 1:
            raise BadRequest(
                "For transform type map the outputpath should be a single column"
            )
        transformed_data = transform_filter(
            transform_func, data, request_data.tolerance
        )

    elif request_data.type == "split":
        if request_data.outputpath is None:
            raise BadRequest(
                "transform type split needs to have outputpath defined in the request body"
            )
        transformed_data = transform_split(transform_func, data, request_data.tolerance)

    elif request_data.type == "concatenate":
        transformed_data = transform_concatenate(
            transform_func, data, request_data.tolerance
        )

    return jsonify(transformed_data)



import os

from dependency_injector.wiring import Provide, inject
from flask import jsonify
from gena import generate_api, generate_app, generate_readonly_api_4dict
from sm.misc.funcs import identity_func
from werkzeug.exceptions import HTTPException

import sand.deserializer as sand_deser
import sand.serializer as sand_ser
from sand.controllers.assistant import assistant_bp
from sand.controllers.project import project_bp
from sand.controllers.search import search_bp
from sand.controllers.settings import setting_bp
from sand.controllers.table import table_bp, table_row_bp
from sand.controllers.transformation import transformation_bp
from sand.helpers.namespace import NamespaceService
from sand.models import EntityAR, SemanticModel
from sand.models.ontology import OntClassAR, OntPropertyAR


@inject
def get_flask_app(
    namespace: NamespaceService = Provide["namespace"],
    entities: EntityAR = Provide["entities"],
    classes: OntClassAR = Provide["classes"],
    properties: OntPropertyAR = Provide["properties"],
):
    app = generate_app(
        [
            table_bp,
            project_bp,
            assistant_bp,
            table_row_bp,
            setting_bp,
            search_bp,
            transformation_bp,
            generate_api(
                SemanticModel,
                deserializers={"data": sand_deser.deserialize_graph},
                batch_serialize=sand_ser.batch_serialize_sms,
            ),
            generate_readonly_api_4dict(
                "entities",
                serialize=sand_ser.serialize_entity,
                id2ent=entities,
            ),
            generate_readonly_api_4dict(
                "classes",
                serialize=sand_ser.serialize_class,
                id2ent=classes,
                unique_field_funcs={"uri": namespace.uri_to_id},
            ),
            generate_readonly_api_4dict(
                "properties",
                serialize=sand_ser.serialize_property,
                id2ent=properties,
                unique_field_funcs={"uri": namespace.uri_to_id},
            ),
        ],
        os.path.dirname(__file__),
    )

    @app.errorhandler(HTTPException)
    def handle_exception(e):
        """Return JSON instead of HTML for HTTP errors."""
        return (
            jsonify(
                {
                    "status": "error",
                    "message": str(e),
                }
            ),
            e.code,
        )

    app.config["MAX_CONTENT_LENGTH"] = 16 * 1024 * 1024  # maximum upload of 16 MB
    return app



from typing import List, Union
from sand.extension_interface.search import IEntitySearch, IOntologySearch, SearchResult


class AggregatedSearch(IEntitySearch, IOntologySearch):

    def __init__(self):
        self.search_entity_objs = []
        self.search_ontology_objs = []

    def add(self, search_obj: Union[IEntitySearch, IOntologySearch]):
        """adds a search object to the aggregated search"""
        if isinstance(search_obj, IEntitySearch):
            self.search_entity_objs.append(search_obj)
        if isinstance(search_obj, IOntologySearch):
            self.search_ontology_objs.append(search_obj)

    def find_class_by_name(self, search_text: str) -> List[SearchResult]:
        """Performs generic class search based on name"""
        payload_results = []
        for search_obj in self.search_ontology_objs:
            payload_results += search_obj.find_class_by_name(search_text)
        return payload_results

    def find_entity_by_name(self, search_text: str) -> List[SearchResult]:
        """Performs generic entity search based on name"""
        payload_results = []
        for search_obj in self.search_entity_objs:
            payload_results += search_obj.find_entity_by_name(search_text)
        return payload_results

    def find_props_by_name(self, search_text: str) -> List[SearchResult]:
        """Performs generic property search based on name"""
        payload_results = []
        for search_obj in self.search_ontology_objs:
            payload_results += search_obj.find_props_by_name(search_text)
        return payload_results



from sand.models.base import db, init_db
from sand.models.entity import Value, EntityAR
from sand.models.project import Project
from sand.models.semantic_model import SemanticModel
from sand.models.table import Table, TableRow, Link, ContextPage
from sand.models.transformation import Transformation

all_tables = [Project, SemanticModel, Table, TableRow, Transformation]



import threading

from dependency_injector.wiring import Provide, inject
from flask import jsonify, request
from flask.blueprints import Blueprint
from gena.serializer import get_dataclass_serializer
from werkzeug.exceptions import BadRequest

from sand.extension_interface.search import IEntitySearch, IOntologySearch, SearchResult

search_bp = Blueprint("search", "search")

GetSearchCache = threading.local()
serializer = get_dataclass_serializer(SearchResult)


@search_bp.route(f"/{search_bp.name}/classes", methods=["GET"])
@inject
def search_classes(ontology_search: IOntologySearch = Provide["ontology_search"]):
    """API Route to search for classes with their names"""
    if "q" not in request.args:
        raise BadRequest("Missing search text")
    query = request.args["q"]

    search_results = ontology_search.find_class_by_name(query)
    serialized_payload = [serializer(item) for item in search_results]
    return jsonify({"items": serialized_payload})


@search_bp.route(f"/{search_bp.name}/entities", methods=["GET"])
@inject
def search_entities(entity_search: IEntitySearch = Provide["entity_search"]):
    """API Route to search for entities with their names"""
    if "q" not in request.args:
        raise BadRequest("Missing search text")
    query = request.args["q"]

    search_results = entity_search.find_entity_by_name(query)
    serialized_payload = [serializer(item) for item in search_results]
    return jsonify({"items": serialized_payload})


@search_bp.route(
    f"/{search_bp.name}/props",
    methods=["GET"],
)
@inject
def search_props(ontology_search: IOntologySearch = Provide["ontology_search"]):
    """API Route to search for properties with their names"""
    if "q" not in request.args:
        raise BadRequest("Missing search text")
    query = request.args["q"]

    search_results = ontology_search.find_props_by_name(query)
    serialized_payload = [serializer(item) for item in search_results]
    return jsonify({"items": serialized_payload})



from abc import ABC, abstractmethod
from typing import List
from dataclasses import dataclass


@dataclass
class SearchResult:
    """
    Search Result dataclass to save the values of each search result in a search
    """
    label: str
    id: str
    description: str
    uri: str


class IEntitySearch(ABC):
    """ Entity Search Interface to support searches from multiple
        KG datastores.
    """

    @abstractmethod
    def find_entity_by_name(self, search_text: str) -> List[SearchResult]:
        """Search Entity using name"""
        pass


class IOntologySearch(ABC):
    """ Class and Property Ontology Search Interface to support searches from multiple
        KG datastores.
    """

    @abstractmethod
    def find_class_by_name(self, search_text: str) -> List[SearchResult]:
        """Search Class using name"""
        pass

    @abstractmethod
    def find_props_by_name(self, search_text: str) -> List[SearchResult]:
        """Search properties using name"""
        pass



from typing import Optional

import click
from loguru import logger
from peewee import fn
from tornado.httpserver import HTTPServer
from tornado.ioloop import IOLoop
from tornado.wsgi import WSGIContainer

from sand.app import get_flask_app
from sand.commands.load import load_dataset
from sand.container import use_container
from sand.helpers.dependency_injection import use_auto_inject
from sand.models import Project, SemanticModel, Table, TableRow, Transformation
from sand.models import db as dbconn
from sand.models import init_db


@click.command()
@click.option("-d", "--db", required=True, help="sand database file")
def init(db):
    """Init database"""
    init_db(db)
    dbconn.create_tables(
        [Project, Table, TableRow, SemanticModel, Transformation], safe=True
    )
    if Project.select().where(fn.Lower(Project.name) == "default").count() == 0:
        Project(name="Default", description="The default project").save()


@click.command()
@click.option("-d", "--db", required=True, help="sand database file")
@click.option(
    "-c",
    "--config",
    help="Path to the configuration file",
)
@click.option("--wsgi", is_flag=True, help="Whether to use wsgi server")
@click.option("-p", "--port", default=5524, help="Listening port")
@click.option(
    "--certfile", default=None, help="Path to the certificate signing request"
)
@click.option("--keyfile", default=None, help="Path to the key file")
def start(
    db: str,
    config: Optional[str],
    wsgi: bool,
    port: int,
    certfile: str,
    keyfile: str,
):
    init_db(db)

    if certfile is None or keyfile is None:
        ssl_options = None
    else:
        ssl_options = {"certfile": certfile, "keyfile": keyfile}
        assert not wsgi

    with use_container(config) as container:
        with use_auto_inject(container):
            app = get_flask_app()

            if wsgi:
                app.run(host="0.0.0.0", port=port)
            else:
                logger.info("Start server in non-wsgi mode")
                http_server = HTTPServer(WSGIContainer(app), ssl_options=ssl_options)
                http_server.listen(address="0.0.0.0", port=port)
                IOLoop.instance().start()


@click.command()
@click.option("-d", "--db", required=True, help="sand database file")
@click.option("--description", help="Description of the project")
@click.argument("name")
def create(db, description: Optional[str], name: str):
    """Create project if not exist"""
    init_db(db)
    Project(name=name, description=description or name).save()


@click.command(help="Remove a project")
@click.option("-d", "--db", required=True, help="sand database file")
@click.argument("name")
def remove(db, name: str):
    """Create project if not exist"""
    init_db(db)
    Project.get(name=name).delete_instance()


@click.group()
def cli():
    pass


cli.add_command(init)
cli.add_command(start)
cli.add_command(create)
cli.add_command(remove)
cli.add_command(load_dataset)


if __name__ == "__main__":
    cli()



from __future__ import annotations

import csv
import zipfile
from dataclasses import dataclass
from functools import lru_cache
from io import BytesIO, StringIO
from typing import List, Literal, Optional

import sm.outputs.semantic_model as O
from dependency_injector.wiring import Provide, inject
from flask import jsonify, make_response, request
from gena import generate_api
from gena.deserializer import (
    generate_deserializer,
    get_dataclass_deserializer,
    get_deserializer_from_type,
)
from peewee import DoesNotExist, fn
from slugify import slugify
from sm.misc.funcs import import_func
from werkzeug.exceptions import BadRequest, NotFound

from sand.config import AppConfig
from sand.deserializer import deser_context_tree
from sand.extension_interface.export import IExport, OutputFormat
from sand.helpers.namespace import NamespaceService
from sand.helpers.service_provider import MultiServiceProvider
from sand.models import SemanticModel, Table, TableRow
from sand.models.ontology import OntClassAR, OntPropertyAR
from sand.models.table import Link

table_bp = generate_api(
    Table,
    deserializers=dict(
        context_tree=deser_context_tree,
        **generate_deserializer(Table, known_field_deserializers={"context_tree"}),
    ),
)


@table_bp.route(
    f"/{table_bp.name}/<id>/export-semantic-models",
    methods=["GET"],
)
@inject
def export_sms(
    id: int,
    ontclass_ar: OntClassAR = Provide["classes"],
    ontprop_ar: OntPropertyAR = Provide["properties"],
):
    subquery = (
        SemanticModel.select(
            SemanticModel.id, fn.MAX(SemanticModel.version).alias("version")
        )
        .where(SemanticModel.table == id)
        .group_by(SemanticModel.table, SemanticModel.name)
        .alias("q1")
    )

    query = (
        SemanticModel.select()
        .where(SemanticModel.table == id)
        .join(subquery, on=(SemanticModel.id == subquery.c.id))
    )

    sms: List[O.SemanticModel] = [r.data for r in query]
    for sm in sms:
        for n in sm.iter_nodes():
            if isinstance(n, O.ClassNode):
                if n.readable_label is None:
                    n.readable_label = (
                        tmp.readable_label
                        if (tmp := ontclass_ar.get_by_uri(n.abs_uri)) is not None
                        else None
                    )
        for e in sm.iter_edges():
            if e.readable_label is None:
                if e.readable_label is None:
                    e.readable_label = (
                        tmp.readable_label
                        if (tmp := ontprop_ar.get_by_uri(e.abs_uri)) is not None
                        else None
                    )

    resp = jsonify([sm.to_dict() for sm in sms])
    if request.args.get("attachment", "false") == "true":
        resp.headers["Content-Disposition"] = (
            f"attachment; filename=semantic-models.json"
        )
    return resp


@table_bp.route(
    f"/{table_bp.name}/<id>/export-linked-entities",
    methods=["GET"],
)
@inject
def export_linked_entities(
    id: int, export: MultiServiceProvider[IExport] = Provide["export"]
):
    table: Table = Table.get_by_id(id)
    rows: List[TableRow] = list(TableRow.select().where(TableRow.table == table))

    output = []
    for ri, row in enumerate(rows):
        for ci, col in enumerate(row.row):
            links = row.links.get(str(ci), [])
            for link in links:
                output.append(
                    {
                        "row": ri,
                        "col": ci,
                        "start": link.start,
                        "end": link.end,
                        "url": link.url,
                        "entity": link.entity_id,
                    }
                )
    f = StringIO()
    writer = csv.writer(
        f, delimiter=",", quoting=csv.QUOTE_MINIMAL, lineterminator="\n"
    )
    header = ["row", "col", "start", "end", "url", "entity"]
    writer.writerow(header)
    for record in output:
        writer.writerow([record[h] for h in header])
    resp = make_response(f.getvalue())
    resp.headers["Content-Type"] = "text/csv; charset=utf-8"
    if request.args.get("attachment", "false") == "true":
        resp.headers["Content-Disposition"] = (
            f"attachment; filename={get_friendly_fs_name(str(table.name))}.csv"
        )
    return resp


@table_bp.route(
    f"/{table_bp.name}/<id>/export-full-model",
    methods=["GET"],
)
@inject
def export_full_model(
    id: int, export: MultiServiceProvider[IExport] = Provide["export"]
):
    # load table
    table: Table = Table.get_by_id(id)

    # load models
    subquery = (
        SemanticModel.select(
            SemanticModel.id, fn.MAX(SemanticModel.version).alias("version")
        )
        .where(SemanticModel.table == table)
        .group_by(SemanticModel.table, SemanticModel.name)
        .alias("q1")
    )
    query = (
        SemanticModel.select()
        .where(SemanticModel.table == table)
        .join(subquery, on=(SemanticModel.id == subquery.c.id))
    )
    sms: List[SemanticModel] = list(query)

    if len(sms) == 0:
        raise BadRequest("Exporting data requires the table to be modeled")

    if "sm" in request.args:
        sms = [sm for sm in sms if sm.name == request.args["sm"]]
        if len(sms) == 0:
            raise BadRequest(
                f"The semantic model with name {request.args['sm']} is not found"
            )
    else:
        if len(sms) != 1:
            raise BadRequest(
                "There are more than one semantic model for this table. Please specify the semantic model you want to export via the 'sm' query parameter"
            )
    sm = sms[0]
    rows: List[TableRow] = list(TableRow.select().where(TableRow.table == table))
    export_obj = export.get_default()
    datamodel = export_obj.export_data_model(table, sm.data)
    resources = export_obj.export_extra_resources(table, rows, sm.data)

    # make a zip file and return it
    # output = StringIO()
    output = BytesIO()
    with zipfile.ZipFile(output, "w") as zip_file:
        for name, content in datamodel.items():
            with zip_file.open(name, "w") as f:
                f.write(content.encode())
        for res_name, res_content in resources.items():
            assert f"{res_name}.txt" not in datamodel
            with zip_file.open(f"{res_name}.txt", "w") as f:
                f.write(res_content.encode())
    resp = make_response(output.getvalue())
    resp.headers["Content-Type"] = "application/zip; charset=utf-8"
    resp.headers["Content-Disposition"] = (
        f"attachment; filename={get_friendly_fs_name(str(table.name))}.zip"
    )
    return resp


@table_bp.route(
    f"/{table_bp.name}/<id>/export",
    methods=["GET"],
)
@inject
def export_table_data(
    id: int, export: MultiServiceProvider[IExport] = Provide["export"]
):
    # load table
    table: Table = Table.get_by_id(id)

    # load models
    subquery = (
        SemanticModel.select(
            SemanticModel.id, fn.MAX(SemanticModel.version).alias("version")
        )
        .where(SemanticModel.table == table)
        .group_by(SemanticModel.table, SemanticModel.name)
        .alias("q1")
    )
    query = (
        SemanticModel.select()
        .where(SemanticModel.table == table)
        .join(subquery, on=(SemanticModel.id == subquery.c.id))
    )
    sms: List[SemanticModel] = list(query)

    if len(sms) == 0:
        raise BadRequest("Exporting data requires the table to be modeled")

    if "sm" in request.args:
        sms = [sm for sm in sms if sm.name == request.args["sm"]]
        if len(sms) == 0:
            raise BadRequest(
                f"The semantic model with name {request.args['sm']} is not found"
            )
    else:
        if len(sms) != 1:
            raise BadRequest(
                "There are more than one semantic model for this table. Please specify the semantic model you want to export via the 'sm' query parameter"
            )
    sm = sms[0]

    # load rows
    rows: List[TableRow] = list(TableRow.select().where(TableRow.table == table))

    try:
        content = export.get_default().export_data(
            table, rows, sm.data, OutputFormat.TTL
        )
    except Exception as e:
        raise BadRequest(f"Failed to export the table. Reason: {str(e)}")

    resp = make_response(content)
    resp.headers["Content-Type"] = "text/ttl; charset=utf-8"
    if request.args.get("attachment", "false") == "true":
        resp.headers["Content-Disposition"] = (
            f"attachment; filename={get_friendly_fs_name(str(table.name))}.ttl"
        )
    return resp


@lru_cache
@inject
def get_export(
    name: Literal["default"] | str, appcfg: AppConfig = Provide["appcfg"]
) -> IExport:
    return import_func(appcfg.export.get_func(name))()


table_row_bp = generate_api(TableRow)
deser_list_links = get_deserializer_from_type(List[Link], {})
assert deser_list_links is not None


@dataclass
class UpdateColumnLinksInput:
    table: int
    column: int
    text: str
    entity_id: Optional[str]


deser_update_column_links = get_dataclass_deserializer(UpdateColumnLinksInput, {})
assert deser_update_column_links is not None


@table_row_bp.route(f"/{table_row_bp.name}/<id>/cells/<column>", methods=["PUT"])
def update_cell_link(id: int, column: int):
    """
    Update the link entity & candidate entities for a given table row and column.
    """
    try:
        row: TableRow = TableRow.get_by_id(id)
    except DoesNotExist:
        raise NotFound(f"Record {id} does not exist")

    try:
        column = int(column)
        if column >= len(row.row):
            raise BadRequest(f"Invalid column {column} value")
    except ValueError:
        raise BadRequest(f"Column {column} is not an integer")

    request_json = request.get_json()
    if request_json is None:
        raise BadRequest("Missing request body")

    if "links" not in request_json:
        raise KeyError(f"Field 'links' is required")
    if not isinstance(request_json["links"], list) or not all(
        isinstance(link, dict) for link in request_json["links"]
    ):
        raise KeyError(f"Field 'links' must be a list of dictionaries")

    if (
        len(request_json["links"]) > 0
        and "candidate_entities" not in request_json["links"][0]
    ):
        # add back the candidate so we can deserialize them
        for link in request_json["links"]:
            link["candidate_entities"] = []
        links: List[Link] = deser_list_links(request_json["links"])

        # then, we just need to update the links' entities
        if len(row.links.get(str(column), [])) == 0:
            row.links[str(column)] = links
        else:
            if len(row.links[str(column)]) != len(links):
                raise BadRequest(
                    "Number of links in request does not match number of links in database"
                )
            for i, link in enumerate(links):
                db_link = row.links[str(column)][i]
                db_link.start = link.start
                db_link.end = link.end
                db_link.entity_id = link.entity_id
    else:
        row.links[str(column)] = deser_list_links(request_json["links"])

    row.save()
    return jsonify({"success": True})


@table_row_bp.route(f"/{table_row_bp.name}/update_column_links", methods=["PUT"])
def update_column_links():
    request_json = request.get_json()
    if request_json is None:
        raise BadRequest("Missing request body")

    args: UpdateColumnLinksInput = deser_update_column_links(request_json)
    table: Table = Table.get_by_id(args.table)
    if args.column >= len(table.columns):
        raise BadRequest(f"Invalid column {args.column} value")

    rows: List[TableRow] = list(TableRow.select().where(TableRow.table == table))
    str_column = str(args.column)
    for row in rows:
        if str(row.row[args.column]) != args.text:
            continue

        if str_column not in row.links:
            row.links[str_column] = [
                Link(
                    start=0,
                    end=len(str(row.row[args.column])),
                    url=None,
                    entity_id=args.entity_id,
                    candidate_entities=[],
                )
            ]
        else:
            db_link = row.links[str_column][0]
            db_link.start = 0
            db_link.end = len(str(row.row[args.column]))
            db_link.entity_id = args.entity_id

        row.save()
    return jsonify({"success": True})


def get_friendly_fs_name(name: str) -> str:
    return slugify(name.replace("/", "_"), lowercase=False).replace("-", "_")



from __future__ import annotations

import os
from dataclasses import dataclass
from functools import cached_property
from pathlib import Path
from typing import Literal, TypedDict

import serde.yaml
from sm.misc.funcs import import_attr
from sm.namespaces.namespace import KnowledgeGraphNamespace

_ROOT_DIR = Path(os.path.abspath(__file__)).parent.parent
PACKAGE_DIR = str(Path(os.path.abspath(__file__)).parent)
FROM_SITEPACKAGES = _ROOT_DIR.name == "site-packages"

CACHE_SIZE = 10240
# indicate that this string is a relative path.
RELPATH_CONST = "::RELPATH::"


@dataclass
class AppConfig:
    kgns: str
    entity: EntityConfig
    clazz: OntResourceConfig
    property: OntResourceConfig
    semantic_model: SemanticModelConfig
    assistant: FnConfig
    search: SearchConfig
    export: FnConfig

    @staticmethod
    def from_yaml(infile: Path | str) -> AppConfig:
        cwd = Path(infile).parent
        obj = serde.yaml.deser(infile)

        return AppConfig(
            kgns=obj["kgns"],
            entity=EntityConfig(
                constructor=obj["entity"]["constructor"],
                args=AppConfig._parse_args(obj["entity"]["args"], cwd),
                default=obj["entity"]["default"],
                instanceof=obj["entity"]["instanceof"],
                nil=IdAndUri(obj["entity"]["nil"]["id"], obj["entity"]["nil"]["uri"]),
                new_entity_template=obj["entity"]["new_entity_template"],
            ),
            clazz=OntResourceConfig(
                constructor=obj["class"]["constructor"],
                args=AppConfig._parse_args(obj["class"]["args"], cwd),
                default=obj["class"]["default"],
            ),
            property=OntResourceConfig(
                constructor=obj["property"]["constructor"],
                args=AppConfig._parse_args(obj["property"]["args"], cwd),
                default=obj["property"]["default"],
            ),
            semantic_model=SemanticModelConfig(
                identifiers=obj["semantic_model"]["identifiers"],
                statements=obj["semantic_model"]["statements"],
            ),
            search=SearchConfig(
                entity=obj["search"]["entity"],
                ontology=obj["search"]["ontology"],
            ),
            assistant=FnConfig(
                default=obj["assistant"].pop("default"),
                funcs=AppConfig._parse_args(obj["assistant"], cwd),
            ),
            export=FnConfig(default=obj["export"].pop("default"), funcs=obj["export"]),
        )

    @staticmethod
    def default() -> AppConfig:
        return AppConfig.from_yaml(os.path.join(PACKAGE_DIR, "config.default.yml"))

    def update(self, obj: AppConfig):
        for k, v in obj.__dict__.items():
            if k.startswith("_"):
                continue
            self.__dict__[k] = v

    @staticmethod
    def _parse_args(obj: dict, cwd: Path):
        out = {}
        for k, v in obj.items():
            if isinstance(v, str) and v.startswith(RELPATH_CONST):
                out[k] = str(cwd / v[len(RELPATH_CONST) :])
            elif isinstance(v, dict):
                out[k] = AppConfig._parse_args(v, cwd)
            else:
                out[k] = v
        return out


@dataclass
class SearchConfig:
    entity: str
    ontology: str


class FnConstructor(TypedDict):
    constructor: str
    args: dict


@dataclass
class FnConfig:
    default: str
    funcs: dict[str, str | FnConstructor]

    def get_func(self, name: Literal["default"] | str) -> str | FnConstructor:
        if name == "default":
            return self.funcs[self.default]
        return self.funcs[name]


@dataclass
class EntityConfig:
    constructor: str
    args: dict
    default: str
    instanceof: dict[str, str]
    nil: IdAndUri
    new_entity_template: str


@dataclass
class IdAndUri:
    id: str
    uri: str


@dataclass
class OntResourceConfig:
    constructor: str
    args: dict
    default: str


@dataclass
class SemanticModelConfig:
    identifiers: list[str]
    statements: list[str]

    @cached_property
    def identifiers_set(self) -> set[str]:
        return set(self.identifiers)



from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING, Literal, Mapping, Union

from dependency_injector.wiring import Provide, inject
from kgdata.models.multilingual import MultiLingualString, MultiLingualStringList
from kgdata.wikidata.models.wdvalue import (
    ValueGlobeCoordinate,
    ValueMonolingualText,
    ValueQuantity,
    ValueTime,
)
from sm.misc.funcs import import_func

from sand.config import AppConfig
from sand.helpers.mapping_utils import KGMapping

if TYPE_CHECKING:
    from sand.helpers.namespace import NamespaceService


@dataclass
class Entity:
    id: str
    uri: str
    label: MultiLingualString
    aliases: MultiLingualStringList
    description: MultiLingualString
    properties: dict[str, list[Statement]]

    @property
    def readable_label(self):
        return self.label

    @property
    def instanceof(self):
        """Get the property representing the instanceof relation."""
        raise NotImplementedError()


@dataclass
class Statement:
    __slots__ = ("value", "qualifiers", "qualifiers_order")
    value: Value
    qualifiers: dict[str, list[Value]]
    # list of qualifiers id that records the order (as dict lacks of order)
    qualifiers_order: list[str]


@dataclass
class Value:
    __slots__ = ("type", "value")
    type: Literal[
        "string", "time", "quantity", "globecoordinate", "entityid", "monolingualtext"
    ]
    value: Union[
        str,
        ValueGlobeCoordinate,
        ValueQuantity,
        ValueTime,
        ValueMonolingualText,
    ]

    def is_entity_id(self):
        return self.type == "entityid"


@inject
def get_default_entities(appcfg: AppConfig = Provide["appcfg"]):
    mapping = {
        appcfg.entity.nil.id: Entity(
            id=appcfg.entity.nil.id,
            uri=appcfg.entity.nil.uri,
            label=MultiLingualString.en("NIL"),
            aliases=MultiLingualStringList(lang2values={"en": []}, lang="en"),
            description=MultiLingualString.en("the correct entity is absent in KG"),
            properties={},
        )
    }
    mapping.update(import_func(appcfg.entity.default)())
    return mapping


class EntityAR(KGMapping[Entity]):
    @staticmethod
    @inject
    def init(
        appcfg: AppConfig = Provide["appcfg"],
        namespace: NamespaceService = Provide["namespace"],
        default_entities: Mapping[str, Entity] = Provide["default_entities"],
    ):
        return EntityAR(
            import_func(appcfg.entity.constructor)(**appcfg.entity.args),
            default_entities,
            namespace.uri_to_id,
        )



from __future__ import annotations

from pathlib import Path
from typing import List, Optional, Tuple

import click
from dependency_injector.wiring import Provide, inject
from sm.dataset import Dataset, Example, FullTable
from sm.misc.funcs import import_func
from sm.outputs.semantic_model import ClassNode, DataNode
from tqdm.auto import tqdm

from sand.container import use_container
from sand.helpers.dependency_injection import use_auto_inject
from sand.models import (
    ContextPage,
    Link,
    Project,
    SemanticModel,
    Table,
    TableRow,
    Value,
)
from sand.models import db as dbconn
from sand.models import init_db
from sand.models.ontology import OntClassAR, OntPropertyAR


@click.command(name="load")
@click.option("-d", "--db", required=True, help="smc database file")
@click.option(
    "-c",
    "--config",
    help="Path to the configuration file",
)
@click.option("-p", "--project", default="default", help="Project name")
@click.option(
    "--dataset",
    required=True,
    help="Path to tables or a python function that returns the dataset in the following format: <python_func>::<dataset_name>",
)
@click.option(
    "-n",
    "--n-tables",
    default=-1,
    type=int,
    help="Number of tables to load (negative number or zero to load all)",
)
@click.option(
    "--add-missing-readable-label",
    is_flag=True,
    help="Attempt to add readable label for nodes and edges that don't have them",
)
def load_dataset(
    db: str,
    config: Optional[str],
    project: str,
    dataset: str,
    n_tables: int,
    add_missing_readable_label: bool,
):
    """Load a dataset into a project"""
    init_db(db)

    if dataset.find("::") != -1:
        func, dsquery = dataset.split("::")
        examples = import_func(func)(dsquery).load()
    else:
        examples = Dataset(Path(dataset)).load()

    if n_tables > 0:
        examples = examples[:n_tables]

    with use_container(config) as container:
        with use_auto_inject(container):
            import_examples(
                project,
                examples,
                add_missing_readable_label,
            )


@inject
def import_examples(
    project: str,
    examples: list[Example[FullTable]],
    add_missing_readable_label: bool,
    ontclass_ar: OntClassAR = Provide["classes"],
    ontprop_ar: OntPropertyAR = Provide["properties"],
):
    if add_missing_readable_label:
        for ex in tqdm(examples, "add missing readable labels"):
            for sm in ex.sms:
                for n in sm.iter_nodes():
                    if isinstance(n, ClassNode):
                        if n.readable_label is None:
                            n.readable_label = (
                                tmp.readable_label
                                if (tmp := ontclass_ar.get_by_uri(n.abs_uri))
                                is not None
                                else None
                            )
                for e in sm.iter_edges():
                    if e.readable_label is None:
                        if e.readable_label is None:
                            e.readable_label = (
                                tmp.readable_label
                                if (tmp := ontprop_ar.get_by_uri(e.abs_uri)) is not None
                                else None
                            )

    with dbconn:
        p = Project.get(name=project)
        for e in tqdm(examples, desc="Loading examples"):
            save_example(p, e)


def save_example(project: Project, example: Example[FullTable]):
    mtbl, mrows = convert_linked_table(example.table)
    mtbl.project = project
    mtbl.save()

    for row in mrows:
        row.save()

    for i, sm in enumerate(example.sms):
        # make sure that the semantic model has all columns in the table
        newsm = sm.deep_copy()
        for col in example.table.table.columns:
            if not newsm.has_data_node(col.index):
                newsm.add_node(
                    DataNode(col_index=col.index, label=col.clean_multiline_name or "")
                )

        SemanticModel(
            table=mtbl,
            name=f"sm-{i}",
            description="",
            version=1,
            data=newsm,
        ).save()


def convert_linked_table(tbl: FullTable) -> Tuple[Table, List[TableRow]]:
    """Convert linked table to smc's tables and rows.
    Note that the project of the table is not set, which is supposed to be set later.
    """
    context_values = [
        Value("entityid", str(entityid)) for entityid in tbl.context.entities
    ]

    if tbl.context.page_url is not None:
        context_page = ContextPage(
            tbl.context.page_url,
            tbl.context.page_title or "",
            (str(tbl.context.entities[0]) if len(tbl.context.entities) > 0 else None),
        )
    else:
        context_page = None

    nrows, ncols = tbl.table.shape()
    links = {}
    for ri in range(nrows):
        links[ri] = {}

    for ri, ci, lst in tbl.links.enumerate_flat_iter():
        if len(lst) == 0:
            continue
        links[ri][ci] = [
            Link(
                start=l.start,
                end=l.end,
                url=l.url,
                entity_id=str(l.entities[0]) if len(l.entities) > 0 else None,
                candidate_entities=[],  # do not load the entities because this is filled by methods
            )
            for l in lst
        ]

    columns = [col.name for col in tbl.table.columns]
    mtbl = Table(
        name=tbl.table.table_id,
        description="",
        columns=columns,
        size=nrows,
        context_values=context_values,
        context_tree=[],
        context_page=context_page,
        project=None,
    )
    mrows = []

    for ri in range(nrows):
        row = [tbl.table[ri, ci] for ci in range(len(columns))]
        mrows.append(TableRow(table=mtbl, index=ri, row=row, links=links[ri]))

    return mtbl, mrows



from __future__ import annotations

from dataclasses import dataclass
from typing import Dict, List, Optional

import sm.outputs.semantic_model as O
from dependency_injector.wiring import Provide, inject
from flask import jsonify, request
from flask.blueprints import Blueprint
from gena.deserializer import get_dataclass_deserializer
from peewee import DoesNotExist
from werkzeug.exceptions import BadRequest, NotFound

import sand.serializer as sand_ser
from sand.config import AppConfig
from sand.extension_interface.assistant import IAssistant
from sand.helpers.service_provider import MultiServiceProvider
from sand.helpers.tree_utils import TreeStruct
from sand.models.entity import Entity, EntityAR
from sand.models.ontology import OntClass, OntClassAR
from sand.models.table import Link, Table, TableRow


@dataclass
class GatherColumnTypesInput:
    table: int
    column: int


deser_gather_column_types = get_dataclass_deserializer(GatherColumnTypesInput, {})
assert deser_gather_column_types is not None

assistant_bp = Blueprint("assistant", "assistant")


@assistant_bp.route(
    f"/{assistant_bp.name}/predict/<table_id>",
    methods=["GET"],
)
@inject
def predict_semantic_desc(
    table_id: int,
    assistant_service: MultiServiceProvider[IAssistant] = Provide["assistant"],
    entity_ar: EntityAR = Provide["entities"],
    ontclass_ar: OntClassAR = Provide["classes"],
    ontprop_ar: OntClassAR = Provide["properties"],
):
    table = Table.get_by_id(table_id)
    rows: List[TableRow] = list(
        TableRow.select().where(TableRow.table == table).order_by(TableRow.index)
    )

    selected_assistants = {
        name: assistant_service.get(name)
        for name in request.args.get("algorithm", "").split(",")
        if name in assistant_service.get_available_providers()
    }
    if len(selected_assistants) == 0:
        selected_assistants = {"default": assistant_service.get_default()}

    outputs = {}
    for name, assistant in selected_assistants.items():
        sm, outputrows = assistant.predict(table, rows)
        if sm is not None:
            for node in sm.iter_nodes():
                if isinstance(node, O.ClassNode):
                    if node.readable_label is None:
                        tmpcls = ontclass_ar.get_by_uri(node.abs_uri)
                        if tmpcls is not None:
                            node.readable_label = tmpcls.readable_label
                        else:
                            node.readable_label = node.rel_uri
                elif isinstance(node, O.LiteralNode):
                    if (
                        node.readable_label is None
                        and node.datatype == O.LiteralNodeDataType.Entity
                    ):
                        tmpent = entity_ar.get_by_uri(node.value)
                        if tmpent is not None:
                            node.readable_label = tmpent.readable_label
                        else:
                            node.readable_label = node.value
            for edge in sm.iter_edges():
                if edge.readable_label is None:
                    tmpprop = ontprop_ar.get_by_uri(edge.abs_uri)
                    if tmpprop is not None:
                        edge.readable_label = tmpprop.readable_label
                    else:
                        edge.readable_label = edge.rel_uri

            sm = sand_ser.serialize_graph(sm, columns=None)
        if outputrows is not None:
            # preserved the manual entity linking from the users
            for outputrow, row in zip(outputrows, rows):
                assert outputrow.id == row.id
                for ci, links in outputrow.links.items():
                    if ci in row.links:
                        glinks: Dict[int, List[Link]] = {}
                        for link in row.links[ci]:
                            if link.entity_id is None:
                                continue
                            glinks.setdefault(link.start, []).append(link)
                        for link in links:
                            for glink in glinks.get(link.start, []):
                                if link.end == glink.end:
                                    link.entity_id = glink.entity_id
            outputrows = [row.to_dict() for row in outputrows]

        outputs[name] = {"sm": sm, "rows": outputrows}

    return jsonify(outputs)


@assistant_bp.route(
    f"{assistant_bp.name}/column-types",
    methods=["POST"],
)
@inject
def gather_column_types(
    appcfg: AppConfig = Provide["appcfg"],
    entity_ar: EntityAR = Provide["entities"],
    ontclass_ar: OntClassAR = Provide["classes"],
):
    """Gather types of the column via its entities"""
    request_json = request.get_json()
    if request_json is None:
        raise BadRequest("Missing request body")

    args: GatherColumnTypesInput = deser_gather_column_types(request_json)

    try:
        table: Table = Table.get_by_id(args.table)
    except DoesNotExist:
        raise NotFound(f"Record {id} does not exist")

    if args.column >= len(table.columns):
        raise BadRequest(f"Invalid column {args.column} value")

    rows: List[TableRow] = list(TableRow.select().where(TableRow.table == table))
    column = str(args.column)
    ents: Dict[str, Optional[Entity]] = {}

    nil_entity_id = appcfg.entity.nil.id
    for row in rows:
        for link in row.links.get(column, []):
            if (
                link.entity_id is not None
                and link.entity_id != nil_entity_id
                and link.entity_id not in ents
            ):
                ents[link.entity_id] = entity_ar.get(link.entity_id, None)
            for can_ent in link.candidate_entities:
                ents[can_ent.entity_id] = entity_ar.get(can_ent.entity_id, None)

    classes: Dict[str, OntClass] = {}
    for ent in ents.values():
        if ent is None:
            continue
        for stmt in ent.properties.get(ent.instanceof, []):
            if stmt.value.is_entity_id():
                klass_id = stmt.value.value
                assert isinstance(klass_id, str)
                klass = ontclass_ar.get(klass_id, None)
                if klass is not None:
                    classes[klass_id] = klass

    tree = TreeStruct.construct(classes, get_parents=lambda id: classes[id].parents)
    return jsonify({p: list(cs) for p, cs in tree.p2cs.items()})



from __future__ import annotations

from typing import Generic, Literal, TypeVar

from sm.misc.funcs import import_func

from sand.config import FnConfig

T = TypeVar("T")


class MultiServiceProvider(Generic[T]):
    def __init__(self, cfg: FnConfig):
        self.cfg = cfg
        self.services: dict[str, T] = {}

    def get_default(self) -> T:
        return self.get("default")

    def get(self, name: Literal["default"] | str) -> T:
        if name not in self.services:
            fnconstructor = self.cfg.get_func(name)
            if isinstance(fnconstructor, str):
                fn = import_func(fnconstructor)()
            else:
                fn = import_func(fnconstructor["constructor"])(
                    **fnconstructor.get("args", {})
                )
            self.services[name] = fn
        return self.services[name]

    def get_available_providers(self) -> list[str]:
        return list(self.cfg.funcs.keys())


def get_service(path: str):
    return import_func(path)()



from __future__ import annotations

import re
from typing import List, Mapping, TypeVar

from dependency_injector.wiring import Provide, inject
from sand.extension_interface.search import IEntitySearch, IOntologySearch, SearchResult
from sand.models.entity import Entity
from sand.models.ontology import OntClass, OntProperty

T = TypeVar("T", Entity, OntClass, OntProperty)


class DefaultSearch(IEntitySearch, IOntologySearch):
    def __init__(
        self,
        default_entities: Mapping[str, Entity],
        default_classes: Mapping[str, OntClass],
        default_properties: Mapping[str, OntProperty],
    ):
        self.default_entities = default_entities
        self.default_classes = default_classes
        self.default_properties = default_properties

    @staticmethod
    @inject
    def create(
        default_entities: Mapping[str, Entity] = Provide["default_entities"],
        default_classes: Mapping[str, OntClass] = Provide["default_classes"],
        default_properties: Mapping[str, OntProperty] = Provide["default_properties"],
    ):
        return DefaultSearch(default_entities, default_classes, default_properties)

    def local_search(self, mapping: Mapping[str, T], search_text: str) -> list[T]:
        """performs local partial text search across default entities/classes/properties"""
        query_tokens = re.findall(r"[a-z]+|\d+", search_text.lower())
        search_results = []

        if not query_tokens:
            return search_results

        for object in mapping.values():
            label = object.label.lower()
            if all(token in label for token in query_tokens):
                search_results.append(object)

        return search_results

    def find_class_by_name(self, search_text: str) -> List[SearchResult]:
        """Uses local default classes to search for entities using their name/text."""
        local_search_results = self.local_search(self.default_classes, search_text)
        payload_results = []

        for entity in local_search_results:
            item = SearchResult(
                label=entity.label,
                id=entity.id,
                description=entity.description,
                uri=entity.uri,
            )
            payload_results.append(item)

        return payload_results

    def find_entity_by_name(self, search_text: str) -> List[SearchResult]:
        """Uses local default entities to search for entities using their name/text."""
        local_search_results = self.local_search(self.default_entities, search_text)
        payload_results = []

        for entity in local_search_results:
            item = SearchResult(
                label=entity.label,
                id=entity.id,
                description=entity.description,
                uri=entity.uri,
            )
            payload_results.append(item)

        return payload_results

    def find_props_by_name(self, search_text: str) -> List[SearchResult]:
        """Uses local default properties to search for entities using their name/text."""
        local_search_results = self.local_search(self.default_properties, search_text)
        payload_results = []

        for entity in local_search_results:
            item = SearchResult(
                label=entity.label,
                id=entity.id,
                description=entity.description,
                uri=entity.uri,
            )
            payload_results.append(item)

        return payload_results



import functools
from pathlib import Path
from typing import Any, Callable, Mapping, TypeVar, Union

from peewee import Field, Model, SqliteDatabase
from sand.config import CACHE_SIZE

# TODO: consider moving to APSWDatabase
db = SqliteDatabase(None)


def init_db(dbfile: Union[str, Path]):
    """Initialize database"""
    global db
    db.init(str(dbfile), pragmas={"foreign_keys": 1})


class BaseModel(Model):
    class Meta:
        database = db


class ClassField(Field):
    field_type = "BLOB"

    def __init__(self, cls, **kwargs):
        super().__init__(**kwargs)
        self.db_value = getattr(cls, "db_value")
        self.python_value = getattr(cls, "python_value")


class BlobField(Field):
    field_type = "BLOB"

    def __init__(self, serialize, deserialize, **kwargs):
        super().__init__(**kwargs)
        self.db_value = serialize
        self.python_value = deserialize


K = TypeVar("K")
V = TypeVar("V")


class StoreWrapper(Mapping[K, V]):
    def __init__(
        self,
        store: Mapping[K, V],
        key_deser: Callable[[K], Any],
        val_deser: Callable[[Any], V],
    ):
        self.store = store
        self.key_deser = key_deser
        self.val_deser = val_deser

    # @functools.lru_cache(maxsize=CACHE_SIZE)
    def __contains__(self, key):
        # print(key)
        # print(self.key_deser(key))
        return self.key_deser(key) in self.store

    # @functools.lru_cache(maxsize=CACHE_SIZE)
    def __getitem__(self, key):
        # print(key)
        # print(self.key_deser(key))
        val = self.store[self.key_deser(key)]
        return self.val_deser(val)

    def __setitem__(self, key, val):
        raise NotImplementedError(
            f"{self.__class__.__name__} does not support __setitem__ function"
        )

    def __delitem__(self, key):
        raise NotImplementedError(
            f"{self.__class__.__name__} does not support __delitem__ function"
        )

    def __len__(self):
        raise NotImplementedError(
            f"{self.__class__.__name__} does not support __len__ function"
        )

    def __iter__(self):
        raise NotImplementedError(
            f"{self.__class__.__name__} does not support __iter__ function"
        )

    def values(self):
        return (self.val_deser(v) for v in self.store.values())

    def get(self, key, default=None):
        if not self.__contains__(key):
            return default
        return self.__getitem__(key)



from dataclasses import asdict
from functools import lru_cache
from pathlib import Path
from typing import Union

import requests
import serde.json
from gena.client import Client as IndividualClient
from peewee import fn
from sm.prelude import M
from tqdm import tqdm

from sand.models import Project, SemanticModel, Table, init_db
from sand.models.table import CandidateEntity, Link


class Client:
    def __init__(
        self,
        url: str = "http://localhost",
    ) -> None:
        self.url = url
        self.projects = IndividualClient(f"{url}/api/project")
        self.tables = IndividualClient(f"{url}/api/table")
        self.table_rows = IndividualClient(f"{url}/api/tablerow")
        self.semantic_models = IndividualClient(f"{url}/api/semanticmodel")

    @lru_cache(maxsize=None)
    def get_project_id(self, name: str) -> int:
        """Get project id by name"""
        project = self.projects.get_one({"name": name, "fields": "id"})
        return project["id"]

    @lru_cache(maxsize=None)
    def get_table_id(self, project_name: str, table_name: str):
        project_id = self.get_project_id(project_name)
        table = self.tables.get_one(
            {"project": project_id, "name": table_name, "fields": "id"}
        )
        return table["id"]

    @lru_cache(maxsize=None)
    def get_table_size(self, table_id: int):
        """Return the number of rows in a table."""
        return self.tables.get_one({"id": table_id, "fields": "size"})["size"]

    def get_table_url(self, table_id: int):
        """Return a URL to browse a table."""
        return f"{self.url}/tables/{table_id}"

    def get_table_url_by_name(self, project_name: str, table_name: str):
        """Return a URL to browse a table by its name."""
        return self.get_table_url(self.get_table_id(project_name, table_name))

    def update_column_links(self, table_id: int, column: int, links: list[list[Link]]):
        """Update all links in a column of a table.

        Args:
            table_id: table id
            column: column index
            links: list of links of each cell. A cell may have more than one link or no links.
        """
        for ri, clinks in enumerate(links):
            rowid = self.table_rows.get_one(
                {"table": table_id, "index": ri, "fields": "id"}
            )["id"]
            resp = requests.put(
                f"{self.table_rows.endpoint}/{rowid}/cells/{column}",
                json={"links": [asdict(l) for l in clinks]},
            )
            self.table_rows.assert_resp(resp)
            assert resp.json()["success"]

    def get_column_links(self, table_id: int, column: int) -> list[list[Link]]:
        """Get annotated links of a column of a table.

        Args:
            table_id: table id
            column: column index
        """
        rows = self.table_rows.get(
            {"table": table_id, "fields": "index,links", "limit": 1000}
        )
        assert len(rows) == self.get_table_size(table_id)
        links = [[] for _ in range(len(rows))]
        for row in rows:
            lst = []
            for item in row["links"].get(str(column), []):
                lst.append(
                    Link(
                        start=item["start"],
                        end=item["end"],
                        url=item["url"],
                        entity_id=item["entity_id"],
                        candidate_entities=[
                            CandidateEntity(c["entity_id"], c["probability"])
                            for c in item["candidate_entities"]
                        ],
                    )
                )
            links[row["index"]] = lst
        return links

    def export(self, project_name: str, outdir: Union[str, Path]):
        """Export a project to a directory."""
        raise NotImplementedError()
        # outdir = Path(outdir)
        # project = Project.get(name=project_name)
        # tables = list(project.tables)

        # for tbl in tqdm(tables):
        #     dname = LinkedTable.get_friendly_fs_id(tbl.name)
        #     subquery = (
        #         SemanticModel.select(
        #             SemanticModel.id, fn.MAX(SemanticModel.version).alias("version")
        #         )
        #         .where(SemanticModel.table == tbl.id)
        #         .group_by(SemanticModel.table, SemanticModel.name)
        #         .alias("q1")
        #     )

        #     query = (
        #         SemanticModel.select()
        #         .where(SemanticModel.table == tbl.id)
        #         .join(subquery, on=(SemanticModel.id == subquery.c.id))
        #     )

        #     sms = list(query)
        #     outfile = outdir / dname / f"version.01.json"
        #     outfile.parent.mkdir(exist_ok=True, parents=True)
        #     serde.json.ser([sm.data.to_dict() for sm in sms], outfile, indent=2)

        # return None



from __future__ import annotations
from typing import Literal, Union, List
from peewee import (
    CharField,
    ForeignKeyField,
    TextField,
    IntegerField,
    BooleanField,
)
from playhouse.sqlite_ext import JSONField

from sand.models.base import BaseModel
from sand.models.table import Table


class Transformation(BaseModel):
    name = CharField()
    table = ForeignKeyField(Table, backref="transformations", on_delete="CASCADE")
    mode = CharField()
    datapath: Union[List[str], str] = JSONField()  # type: ignore
    outputpath: List[str] = JSONField()  # type: ignore
    type = CharField()
    code = TextField()
    on_error: Literal[
        "set_to_blank", "store_error", "keep_original", "abort"
    ] = CharField()
    is_draft = BooleanField()
    order = IntegerField()
    insert_after = ForeignKeyField("self", null=True, on_delete="SET NULL")

    def to_dict(self):
        return {
            "id": self.id,
            "name": self.name,
            "table": self.table_id,
            "type": self.type,
            "mode": self.mode,
            "datapath": self.datapath,
            "outputpath": self.outputpath,
            "code": self.on_error,
            "on_error": self.on_error,
            "is_draft": self.is_draft,
            "order": self.order,
            "insert_after": self.insert_after,
        }
